/* SBN machine simulator. */

#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include "asttypes.h"

Stmt *mem;
int nstmts;

extern int quiet;
extern int dump_before;
extern char *output_only;
extern char *the_filename;

extern char *string_of(int intern_num);
extern int intern_tolower(char *outside);

void dump_decls(void);

int count(StmtList prog)
{
	StmtList p;
	int tmp = 1;
	for (p = prog; p->cdr != NULL; p = p->cdr) { tmp++; }
	return tmp;
}

void flatten(StmtList prog)
{
	StmtList p;
	int tmp = 0;
	for (p = prog; p != NULL; p = p->cdr) { mem[tmp++] = p->car; }
}

int resolve_label(int l)
{
	int i;

	for (i = 0; i < nstmts; i++) {
		if (mem[i]->label == l) {
			return i;
		}
	}
	return -1;
}

int resolve_start_address(void)
{
	int start_sym = intern_tolower("start");
	return resolve_label(start_sym);
}

int resolve_address(Address a)
{
	if (a->absolute != 0) return a->absolute;
	switch(a->type) {
		case AT_LABELREL:
			a->absolute = a->offset + resolve_label(a->label);
			return a->absolute;
			break;
		case AT_PCREL:
			a->absolute = a->offset + a->label;
			return a->absolute;
			break;
		default:
			fprintf(stderr,
				"Internal error: weird address type %d in"
				"resolve_address(0x%lx), aborting\n",a->type,((long)a));
			abort();
	}
}

char *address_tostring(Address a)
{
	char buff[100];
	char signchar;
	int offset;

	if (a->tostring != NULL) { return a->tostring; }

	switch(a->type) {
		case AT_LABELREL:
			if (a->offset != 0) {
				signchar = (a->offset < 0) ? '-' : '+';
				offset = (a->offset < 0) ? -a->offset : a->offset;
				sprintf(buff, "%s%c%d", string_of(a->label), signchar, offset);
			} else {
				sprintf(buff, "%s", string_of(a->label));
			}

			a->tostring = strdup(buff);
			return a->tostring;
			break;
		case AT_PCREL:
			if (a->offset != 0) {
				signchar = (a->offset < 0) ? '-' : '+';
				offset = (a->offset < 0) ? -a->offset : a->offset;
				sprintf(buff, ".%c%d", signchar, offset);
			} else {
				sprintf(buff, ".");
			}

			a->tostring = strdup(buff);
			return a->tostring;
			break;
		case AT_ABSOLUTE:
			sprintf(buff, "%d", a->absolute);

			a->tostring = strdup(buff);
			return a->tostring;
			break;
		default:
			fprintf(stderr,
				"Internal error: weird address type %d in"
				"address_tostring(0x%lx), aborting\n",a->type,((long)a));
			abort();
	}
}

int sbn(int curr_pc, Stmt s)
{
	int l, r, b, lsiz;
	if (!quiet) {
		printf("[%3d]\t", curr_pc);
		if (s->label != 0) {
			printf("%s:", string_of(s->label));
			lsiz = strlen(string_of(s->label)) + 1;
		} else {
			lsiz = 0;
		}
		lsiz = 12 - lsiz;
		while (lsiz-- > 0) { putchar(' '); }
	}
	l = resolve_address(s->lhs); 
	if (l < 0) {
		printf("\nERROR: first operand of stmt #%d is defined nowhere\n",
			curr_pc);
		return nstmts;
	}
	r = resolve_address(s->rhs); 
	if (r < 0) {
		printf("\nERROR: second operand of stmt #%d is defined nowhere\n",
			curr_pc);
		return nstmts;
	}
	b = resolve_address(s->branchdest); 
	if (b < 0) {
		printf("\nERROR: third operand of stmt #%d is defined nowhere\n",
			curr_pc);
		return nstmts;
	}
	if (!quiet)
		printf(" sbn %s, %s, %s", address_tostring(s->lhs),
			address_tostring(s->rhs), address_tostring(s->branchdest));
	/* subtract! */
	mem[l]->value -= mem[r]->value;
	/* and branch if negative! */
	if (mem[l]->value < 0) {
		if (!quiet) printf("\t# took the branch\n");
		return b;
	} else {
		if (!quiet) printf("\t# not taking the branch\n");
		return curr_pc + 1;
	}
};

void simulate(StmtList prog)
{
	int pc = -1;

	/* create & fill mem */
	nstmts = count(prog);
	mem = (Stmt *) calloc(nstmts, sizeof(Stmt));
	assert(mem != NULL);
	flatten(prog);
	if (!quiet) {
		printf("# loaded %d statements from %s\n", nstmts, the_filename);
		printf("#[PC]\tLABEL:       INSTRUCTION\t# COMMENT\n");
	}

	if (dump_before)
		dump_decls();

	/* find start & set pc to it */
	pc = resolve_start_address();
	if (pc == -1) {
		fprintf(stderr, "ERROR: no start symbol\n");
		return;
	}

	/* simulate instructions 'till we fall off the end of the code */
	while (pc < nstmts) {
		if (mem[pc]->type != ST_SBN) {
			printf("ERROR: executing something that's not code, pc=%d\n", pc);
			return;
		}
		pc = sbn(pc, mem[pc]);
	}

	if (!quiet)
		printf("### Program completed!\n");
}

void dump_decls(void)
{
	int i;

	if (output_only == NULL) {
		printf("### Dump of declared input/output variables:\n");
		for (i = 0; i < nstmts; i++) {
			if (mem[i]->type == ST_VARDECL) {
				printf("#\t%s = %d\n",string_of(mem[i]->label),mem[i]->value);
			}
		}
	} else {
		int sym, l;

		sym = intern_tolower(output_only);
		if ((l = resolve_label(sym)) < 0) {
			printf("### -o %s was set, but it wasn't declared, ignoring\n",
				output_only);
			output_only = NULL;
			dump_decls();
		} else {
			if (mem[l]->type == ST_VARDECL) {
				printf("%s = %d\n",string_of(mem[l]->label),mem[l]->value);
			} else {
			    printf("### -o %s was set, but it's not a variable, ignoring\n",
				    output_only);
			    output_only = NULL;
			    dump_decls();
			}
		}
	}
}

